from __future__ import annotations

from typing import Any, Dict, Iterator, Optional, Tuple

__all__ = ("User", "Users", "Leaderboard", "RewardRole", "Rewards")


class _SlotsReprMixin:
    __slots__ = ()

    def __repr__(self) -> str:
        inner = ", ".join(
            (
                f"{k}={v!r}"
                for k, v in self.get_slotted_items()
                if v is not None and not k.startswith("_")
            )
        )
        return f"{self.__class__.__name__}({inner})"

    def get_slotted_items(self) -> Iterator[Tuple[str, Any]]:
        for slot in self.__slots__:
            yield slot, getattr(self, slot)


class User(_SlotsReprMixin):
    """
    An Amari user.

    Attributes
    ----------
    guild_id: int
        The user's guild ID.
    user_id: int
        The user's ID.
    name: str
        The user's Discord username. This may not be up to date.
    exp: int
        The user's experience points.
    level: Optional[int]
        The user's level.
    weeklyexp: Optional[int]
        The user's weekly experience points.
    position: Optional[int]
        The user's position in the leaderboard.
    leaderboard: Optional[Leaderboard]
        The leaderboard object the user is in, if a leaderboard endpoint was fetched.
    """

    __slots__ = (
        "user_id",
        "name",
        "guild_id",
        "exp",
        "level",
        "weeklyexp",
        "position",
        "leaderboard",
    )

    def __init__(
        self,
        guild_id: int,
        data: dict,
        position: Optional[int] = None,
        *,
        leaderboard: Optional[Leaderboard] = None,
    ):
        self.guild_id: int = guild_id
        self.user_id: int = int(data["id"])
        self.name: str = data["username"]
        self.exp: int = int(data["exp"])
        self.level: Optional[int] = data.get("level")
        self.weeklyexp: Optional[int] = (
            int(data.get("weeklyExp")) if data.get("weeklyExp") is not None else None
        )
        self.position: Optional[int] = position
        self.leaderboard: Optional[Leaderboard] = leaderboard


class Users:
    """
    An object which holds multiple users.

    Attributes
    ----------
    guild_id: int
        The guild ID.
    users: Dict[int, User]
        The list of users.
    total_members: int
        The amount of users in the list.
    queried_members: int
        The amount of users requested.
    """

    __slots__ = ("guild_id", "users", "total_members", "queried_members")

    def __init__(self, guild_id: int, data: dict):
        self.guild_id: int = guild_id
        self.users: Dict[int, User] = {
            int(user["id"]): User(guild_id, user) for user in data["members"]
        }
        self.total_members: int = data["total_members"]
        self.queried_members: int = data["queried_members"]

    def __repr__(self) -> str:
        return f"<Users guild_id={self.guild_id} user_count={self.total_members}>"

    def __len__(self) -> int:
        return self.total_members

    def __iter__(self) -> Iterator[User]:
        yield from self.users.copy().values()

    def get_user(self, user_id: int, /) -> Optional[User]:
        """
        Get a user from the users object.

        Parameters
        ----------
        user_id: int
            The user's ID.

        Returns
        -------
        Optional[User]
            The user, if found in the users.
        """
        return self.users.get(user_id)

    def add_user(self, user: User, /) -> Users:
        """
        Add a user to the users.

        Parameters
        ----------
        user: User
            The user to add.

        Returns
        -------
        Users
            The object of users the user was added to, for fluent class chaining.
        """
        self.users[user.user_id] = user
        return self


class Leaderboard:
    """
    An Amari leaderboard.

    Attributes
    ----------
    guild_id: int
        The guild ID.
    user_count: int
        The number of users in the leaderboard.
    total_count: Optional[int]
        The total number of users on Amari's API in this leaderboard.
    users: Dict[int, User]
        The users in the leaderboard.
    """

    __slots__ = ("guild_id", "user_count", "total_count", "users")

    def __init__(self, guild_id: int, data: dict):
        self.guild_id: int = guild_id
        self.user_count: int = data["count"]
        self.total_count: Optional[int] = data.get("total_count")
        self.users: Dict[int, User] = {
            int(user_data["id"]): User(guild_id, user_data, i, leaderboard=self)
            for i, user_data in enumerate(data["data"])
        }

    def __repr__(self) -> str:
        return f"<Leaderboard guild_id={self.guild_id} user_count={self.user_count}>"

    def __len__(self) -> int:
        return self.user_count

    def __iter__(self) -> Iterator[User]:
        yield from self.users.copy().values()

    def get_user(self, user_id: int, /) -> Optional[User]:
        """
        Get a user from the leaderboard.

        Parameters
        ----------
        user_id: int
            The user's ID.

        Returns
        -------
        Optional[User]
            The user, if found in the leaderboard.
        """
        return self.users.get(user_id)

    def add_user(self, user: User, /) -> Leaderboard:
        """
        Add a user to the leaderboard.

        Parameters
        ----------
        user: User
            The user to add.

        Returns
        -------
        Leaderboard
            The leaderboard the user was added to, for fluent class chaining.
        """
        self.users[user.user_id] = user
        return self


class RewardRole(_SlotsReprMixin):
    """
    An object representing an Amari reward role.

    Attributes
    ----------
    role_id: int
        The role's ID.
    level: int
        The level that a user needs for the role to be awarded to them.
    rewards: Rewards
        The rewards object this role belongs to.
    """

    __slots__ = ("role_id", "level", "rewards")

    def __init__(self, role_id: int, level: int, rewards: Rewards):
        self.role_id: int = role_id
        self.level: int = level
        self.rewards: Rewards = rewards


class Rewards:
    """
    A collection of Amari reward roles.

    Attributes
    ----------
    guild_id: int
        The guild ID.
    reward_count: int
        The number of reward roles.
    roles: Dict[int, RewardRole]
        The guild's reward roles.
    """

    __slots__ = ("guild_id", "reward_count", "roles")

    def __init__(self, guild_id: int, data: dict):
        self.guild_id: int = guild_id
        self.reward_count: int = int(data["count"])
        self.roles: Dict[int, RewardRole] = {}
        for role_data in data["data"]:
            role_id = int(role_data["roleID"])
            self.roles[role_id] = RewardRole(role_id, role_data["level"], self)

    def __repr__(self) -> str:
        return f"<Rewards guild_id={self.guild_id} reward_count={self.reward_count}>"

    def __len__(self) -> int:
        return self.reward_count

    def __iter__(self) -> Iterator[RewardRole]:
        yield from self.roles.copy().values()

    def get_role(self, role_id: int, /) -> Optional[RewardRole]:
        """
        Get a reward role from the rewards.

        Parameters
        ----------
        role_id: int
            The role's ID.

        Returns
        -------
        Optional[RewardRole]
            The role, if found in the rewards.
        """
        return self.roles.get(role_id)



import asyncio
import logging
import time
from typing import Dict, List, Optional

import aiohttp

from .cache import Cache
from .exceptions import (
    AmariServerError,
    HTTPException,
    InvalidToken,
    NotFound,
    RatelimitException,
)
from .objects import Leaderboard, Rewards, User, Users

__all__ = ("AmariClient",)

logger = logging.getLogger(__name__)


class AmariClient:
    """
    The client used to make requests to the Amari API.

    Attributes
    ----------
    useAntiRatelimit: bool
        Whether to use the anti ratelimit or not.
        IT IS VERY UNRECOMMENDED TO DISABLE THIS FEATURE
        AS IT CAN LEAD TO RATELIMITS

    session: aiohttp.ClientSession
        The client session used to make requests to the Amari API.

    max_requests: int
        The number of requests that can be made per second

    cache_ttl: int
        The time to live for cache entries, in seconds.

    cache: Cache
        The cache instance used to store API responses.

    maxbytes: int
        The maximum total size of cached data in bytes.
    """

    BASE_URL = "https://amaribot.com/api/v1/"

    HTTP_response_errors = {
        404: NotFound,
        403: InvalidToken,
        429: RatelimitException,
        500: AmariServerError,
    }

    def __init__(
        self,
        token: str,
        /,
        *,
        useAntirateLimit: bool = True,
        session: Optional[aiohttp.ClientSession] = None,
        max_requests: int = 55,
        cache_ttl: int = 60,
        maxbytes: int = 25 * 1024 * 1024,  # 25 MiB
    ):
        self.session = session or aiohttp.ClientSession()
        self._default_headers = {"Authorization": token}

        self.lock = asyncio.Lock()

        # Anti Ratelimit section
        self.use_anti_ratelimit = useAntirateLimit

        self.requests = []

        self.max_requests = max_requests
        self.request_period = 60
        self.cache = Cache(ttl=cache_ttl, maxbytes=maxbytes)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc) -> None:
        await self.close()

    async def close(self):
        """
        Closes the client resources.

        This must be called once the client is no longer in use.
        """
        await self.session.close()

    async def check_ratelimit(self):
        async with self.lock:
            while len(self.requests) >= self.max_requests:
                self.requests = [
                    request
                    for request in self.requests
                    if time.time() - request < self.request_period
                ]

                if len(self.requests) >= self.max_requests:
                    await self.wait_for_ratelimit_end()

    async def wait_for_ratelimit_end(self):
        wait_amount = self.request_period - (time.time() - min(self.requests))
        logger.warning(
            f"You are about to be ratelimited! Waiting {round(wait_amount)} seconds."
        )

        await asyncio.sleep(wait_amount)

    async def fetch_user(
        self, guild_id: int, user_id: int, cache: bool = False
    ) -> User:
        """
        Fetches a user from the Amari API.

        Parameters
        ----------
        guild_id: int
            The guild ID to fetch the user from.
        user_id: int
            The user's ID.
        cache: bool
            Whether to use caching for this request.

        Returns
        -------
        User
            The user object.
        """
        if cache:
            key = ("fetch_user", guild_id, user_id)
            data = await self.cache.get(key)
            if data:
                return User(guild_id, data)
            else:
                data = await self.request(f"guild/{guild_id}/member/{user_id}")
                await self.cache.set(key, data)
                return User(guild_id, data)
        else:
            data = await self.request(f"guild/{guild_id}/member/{user_id}")
            return User(guild_id, data)

    async def fetch_users(
        self, guild_id: int, user_ids: List[int], cache: bool = False
    ) -> Users:
        """
        Fetches multiple users from the Amari API.

        Parameters
        ----------
        guild_id: int
            The guild ID to fetch the users from.
        user_ids: List[int]
            The IDs of the users you would like to fetch.
        cache: bool
            Whether to use caching for this request.

        Returns
        -------
        Users
            The users object containing the fetched users.
        """
        if cache:
            members = []
            uncached_user_ids = []

            for user_id in user_ids:
                key = ("fetch_user", guild_id, user_id)
                data = await self.cache.get(key)
                if data:
                    members.append(data)
                else:
                    uncached_user_ids.append(user_id)

            if uncached_user_ids:
                converted_user_ids = [str(user_id) for user_id in uncached_user_ids]
                body = {"members": converted_user_ids}
                fetched_data = await self.request(
                    f"guild/{guild_id}/members",
                    method="POST",
                    extra_headers={"Content-Type": "application/json"},
                    json=body,
                )
                for user_data in fetched_data["members"]:
                    user_id = int(user_data["id"])
                    key = ("fetch_user", guild_id, user_id)
                    await self.cache.set(key, user_data)
                    members.append(user_data)

            data = {
                "members": members,
                "total_members": len(members),
                "queried_members": len(user_ids),
            }
            return Users(guild_id, data)
        else:
            converted_user_ids = [str(user_id) for user_id in user_ids]
            body = {"members": converted_user_ids}
            data = await self.request(
                f"guild/{guild_id}/members",
                method="POST",
                extra_headers={"Content-Type": "application/json"},
                json=body,
            )
            return Users(guild_id, data)

    async def fetch_leaderboard(
        self,
        guild_id: int,
        /,
        *,
        weekly: bool = False,
        raw: bool = False,
        page: Optional[int] = None,
        limit: Optional[int] = None,
        cache: bool = False,
    ) -> Leaderboard:
        """
        Fetches a guild's leaderboard from the Amari API.

        Parameters
        ----------
        guild_id: int
            The guild ID to fetch the leaderboard from.
        weekly: bool
            Choose either to fetch the weekly leaderboard or the regular leaderboard.
        raw: bool
            Whether or not to use the raw endpoint. Raw endpoints do not support pagination but
            will return the entire leaderboard.
        page: int
            The leaderboard page to fetch.
        limit: int
            The amount of users to fetch per page.
        cache: bool
            Whether to use caching for this request.

        Returns
        -------
        Leaderboard
            The guild's leaderboard.
        """
        if cache:
            key = ("fetch_leaderboard", guild_id, weekly, raw, page, limit)
            data = await self.cache.get(key)
            if data:
                return Leaderboard(guild_id, data)
        if raw and page:
            raise ValueError("raw endpoints do not support pagination")
        params = {}
        if page is not None:
            params["page"] = page
        if limit is not None:
            params["limit"] = limit
        lb_type = "weekly" if weekly else "leaderboard"
        endpoint = ["guild", lb_type, str(guild_id)]
        if raw:
            endpoint.insert(1, "raw")

        data = await self.request("/".join(endpoint), params=params)
        if cache:
            await self.cache.set(key, data)
        return Leaderboard(guild_id, data)

    async def fetch_full_leaderboard(
        self, guild_id: int, /, *, weekly: bool = False, cache: bool = False
    ) -> Leaderboard:
        """
        Fetches a guild's full leaderboard from the Amari API.

        Parameters
        ----------
        guild_id: int
            The guild ID to fetch the leaderboard from.
        weekly: bool
            Choose either to fetch the weekly leaderboard or the regular leaderboard.
        cache: bool
            Whether to use caching for this request.

        Returns
        -------
        Leaderboard
            The guild's leaderboard.
        """
        if cache:
            key = ("fetch_full_leaderboard", guild_id, weekly)
            data = await self.cache.get(key)
            if data:
                return Leaderboard(guild_id, data)
        lb_type = "weekly" if weekly else "leaderboard"
        data = await self.request(f"guild/raw/{lb_type}/{guild_id}")
        if cache:
            await self.cache.set(key, data)
        return Leaderboard(guild_id, data)

    async def fetch_rewards(
        self, guild_id: int, /, *, page: int = 1, limit: int = 50, cache: bool = False
    ) -> Rewards:
        """
        Fetches a guild's role rewards from the Amari API.

        Parameters
        ----------
        guild_id: int
            The guild ID to fetch the role rewards from.
        page: int
            The rewards page to fetch.
        limit: int
            The amount of rewards to fetch per page.
        cache: bool
            Whether to use caching for this request.

        Returns
        -------
        Rewards
            The guild's role rewards.
        """
        if cache:
            key = ("fetch_rewards", guild_id, page, limit)
            data = await self.cache.get(key)
            if data:
                return Rewards(guild_id, data)
        params = {"page": page, "limit": limit}
        data = await self.request(f"guild/rewards/{guild_id}", params=params)
        if cache:
            await self.cache.set(key, data)
        return Rewards(guild_id, data)

    @classmethod
    async def check_response_for_errors(cls, response: aiohttp.ClientResponse):
        if response.status > 399 or response.status < 200:
            error = cls.HTTP_response_errors.get(response.status, HTTPException)
            try:
                message = (await response.json())["error"]
            except Exception:
                message = await response.text()
            raise error(response, message)

    async def request(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        params: Dict = {},
        json: Dict = {},
        extra_headers: Dict = {},
    ) -> Dict:
        headers = dict(self._default_headers, **extra_headers)

        if self.use_anti_ratelimit:
            await self.check_ratelimit()

        async with self.session.request(
            method=method,
            url=self.BASE_URL + endpoint,
            json=json,
            headers=headers,
            params=params,
        ) as response:
            if self.use_anti_ratelimit:
                self.requests.append(time.time())

            await self.check_response_for_errors(response)

            return await response.json()



import asyncio
import json
import time
from collections import OrderedDict
from typing import Any, Optional, Tuple


class CacheEntry:
    __slots__ = ("data", "timestamp", "size")

    def __init__(self, data: Any, timestamp: float, size: int):
        self.data = data
        self.timestamp = timestamp
        self.size = size


class Cache:
    """
    A simple LRU cache with TTL and size limit.

    Attributes
    ----------
    ttl: int
        Time to live for cache entries, in seconds.
    maxbytes: int
        Maximum total size of cached data in bytes.
    """

    def __init__(self, ttl: int, maxbytes: int = 25 * 1024 * 1024):  # 25 MiB
        self.ttl = ttl
        self.maxbytes = maxbytes
        self.cache: OrderedDict[Tuple, CacheEntry] = OrderedDict()
        self.total_size = 0
        self.lock = asyncio.Lock()

    async def get(self, key: Tuple) -> Optional[Any]:
        async with self.lock:
            self._remove_expired_entries()
            entry = self.cache.get(key)
            if entry:
                if time.time() - entry.timestamp < self.ttl:
                    self.cache.move_to_end(key)
                    return entry.data
                else:
                    self._remove_entry(key)
            return None

    async def set(self, key: Tuple, data: Any):
        async with self.lock:
            size = len(json.dumps(data).encode("utf-8"))
            entry = CacheEntry(data, time.time(), size)
            self.cache[key] = entry
            self.cache.move_to_end(key)
            self.total_size += size
            self._remove_expired_entries()
            await self._enforce_size_limit()

    def _remove_entry(self, key: Tuple):
        entry = self.cache.pop(key, None)
        if entry:
            self.total_size -= entry.size

    def _remove_expired_entries(self):
        current_time = time.time()
        keys_to_remove = [
            key
            for key, entry in self.cache.items()
            if current_time - entry.timestamp >= self.ttl
        ]
        for key in keys_to_remove:
            self._remove_entry(key)

    async def _enforce_size_limit(self):
        while self.total_size > self.maxbytes:
            key, entry = self.cache.popitem(last=False)
            self.total_size -= entry.size
            await asyncio.sleep(0)



from typing import Optional

import aiohttp

__all__ = (
    "AmariException",
    "HTTPException",
    "NotFound",
    "InvalidToken",
    "RatelimitException",
    "AmariServerError",
)


class AmariException(Exception):
    """Base module Exception class."""


class HTTPException(AmariException):
    """Base Exception for HTTP errors."""

    def __init__(self, response: aiohttp.ClientResponse, message: Optional[str] = None):
        self.status: int = response.status
        self.response: aiohttp.ClientResponse = response
        message = f"({self.status}): {message}" if message else f"({self.status})"
        super().__init__(message)


class NotFound(HTTPException):
    """Raised when the guild or user is not found."""

    def __init__(
        self,
        response: aiohttp.ClientResponse,
        message: Optional[str] = "Guild or user was not found.",
    ):
        super().__init__(response, message)


class InvalidToken(HTTPException):
    """Raised when the authentication key is invalid."""

    def __init__(self, response: aiohttp.ClientResponse, message: Optional[str] = None):
        super().__init__(
            response,
            "Please enter a valid authentication key.\n"
            "You can obtain your key at https://amaribot.com/developer/yourid",
        )


class RatelimitException(HTTPException):
    """Raised when ratelimit responses are recieved."""

    def __init__(
        self,
        response: aiohttp.ClientResponse,
        message: Optional[str] = "Slow down! You are being ratelimited!",
    ):
        super().__init__(response, message)


class AmariServerError(HTTPException):
    """Raised when their is an internal error in the Amari servers."""

    def __init__(
        self,
        response: aiohttp.ClientResponse,
        message: Optional[str] = "There was an internal error in the Amari servers.",
    ):
        super().__init__(response, message)
